/* 
 *      HMAC implementation, see rfc 2104.
 *
 * Copyright (C) 2000 Zero-Knowledge[R] Systems, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 * ---
 *
 * derived from sha1glue.c which was modified by Andrew McDonald <andrew@mcdonald.org.uk> 
 * from md5glue by Alan Smithee.
 *
 * from rfc2104:
 *
 *  (1) append zeros to the end of K to create a B byte string
 *      (e.g., if K is of length 20 bytes and B=64, then K will be
 *      appended with 44 zero bytes 0x00)
 *  (2) XOR (bitwise exclusive-OR) the B byte string computed in step
 *      (1) with ipad
 *  (3) append the stream of data 'text' to the B byte string resulting
 *      from step (2)
 *  (4) apply H to the stream generated in step (3)
 *  (5) XOR (bitwise exclusive-OR) the B byte string computed in
 *      step (1) with opad
 *  (6) append the H result from step (4) to the B byte string
 *      resulting from step (5)
 *  (7) apply H to the stream generated in step (6) and output
 *      the result
 */

#include <linux/config.h>
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/types.h>
#include <asm/uaccess.h>
#include <asm/init.h>
#include <linux/crypto.h>

struct digest_implementation * global_H;

#define PAD_LEN_BYTES 64
#define PAD_LEN_INT (PAD_LEN_BYTES / sizeof(unsigned int))

/* go gcc! */
#define INT_WIDTH_CONST(HEX)                                \
({  unsigned int __ret;                                     \
    switch(sizeof(unsigned int)) {                          \
        case 1: __ret = 0x##HEX; break;                       \
        case 2: __ret = 0x##HEX##HEX; break;                    \
        case 4: __ret = 0x##HEX##HEX##HEX##HEX; break;              \
        case 8: __ret = 0x##HEX##HEX##HEX##HEX##HEX##HEX##HEX##HEX; break;  \
    }                                                       \
    __ret;                                                  \
})

#define MAX_MACLEN 20

struct hmac_local {
    struct digest_implementation *H;
    int pad_ready;
    unsigned int pad[PAD_LEN_INT];
    struct digest_context H_ctx;
    struct digest_context pad_ctx;
};

#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define INFO_LEN MAX(SHA1_WORKING_SIZE, MD5_WORKING_SIZE)
#define HMAC_WORKING_SIZE (sizeof(struct hmac_local) + (2*INFO_LEN))

int
hmac_open (struct digest_context *cx)
{
    struct hmac_local *hl = (struct hmac_local *)cx->digest_info;

	if (!cx || !hl)
		return -EINVAL;

    hl->H = global_H; /* see init_hmac */
    hl->pad_ready = 0; 

    hl->H_ctx.di = hl->H;
    hl->pad_ctx.di = hl->H;

    hl->H_ctx.digest_info = (u32 *)((char *)hl + sizeof(struct hmac_local));
    hl->pad_ctx.digest_info = (u32 *)((char *)hl + sizeof(struct hmac_local) + INFO_LEN);

	return 0;
}

/*
 * Overload the first update to pass in the key.  This
 * is pretty dorky, but we need to know the key length
 * when generating the pad.
 *
 * ugh.
 */
int
hmac_update (struct digest_context *cx, u8 *in, int size)
{
    struct hmac_local *hl = (struct hmac_local *)cx->digest_info;
    int i;

	if (!cx || !in || !hl || !hl->H)
		return -EINVAL;

    if(!hl->pad_ready) {
        memset(hl->pad, 0, PAD_LEN_BYTES);

        if(size <= PAD_LEN_BYTES) {
            memcpy(hl->pad, in, size);

        } else {
            /* ok, gross, we have to hash the key */
            hl->H->open(&(hl->pad_ctx));
            hl->H->update(&(hl->pad_ctx), in, size);
            hl->H->close(&(hl->pad_ctx), (u8*)hl->pad);
        }

        for(i = 0 ; i < PAD_LEN_INT ; i++) 
            hl->pad[i] ^= INT_WIDTH_CONST(36);

        hl->H->open(&(hl->H_ctx));
        hl->H->update(&(hl->H_ctx), (u8 *)hl->pad, PAD_LEN_BYTES);

        hl->pad_ready = 1;

        return 0;
    }

    hl->H->update(&(hl->H_ctx), in, size);

	return 0;
}

int
hmac_digest (struct digest_context *cx, u8 *out)
{
    return -EINVAL;
}

int
hmac_close (struct digest_context *cx, u8 *out)
{
    struct hmac_local *hl = (struct hmac_local *)cx->digest_info;
    unsigned char mac[MAX_MACLEN];
    int i;

	if (!cx || !hl || !hl->H)
		return -EINVAL;

    hl->H->close(&(hl->H_ctx), mac);

    for(i = 0 ; i < PAD_LEN_INT ; i++) 
        hl->pad[i] ^= INT_WIDTH_CONST(36) ^ INT_WIDTH_CONST(5c);

    hl->H->open(&(hl->pad_ctx));
    hl->H->update(&(hl->pad_ctx), (u8 *)hl->pad, PAD_LEN_BYTES);
    hl->H->update(&(hl->pad_ctx), mac, MAX_MACLEN);
    hl->H->close(&(hl->pad_ctx), out);

    memset(mac, 0, MAX_MACLEN);
    memset(hl->pad, 0, PAD_LEN_BYTES);
    hl->H = NULL;

    /* memset 0 hl + ctxs? */

	return 0;
}

void 
hmac_lock (void)
{
	MOD_INC_USE_COUNT;
}

void 
hmac_unlock(void)
{
	MOD_DEC_USE_COUNT;
}

/* This api is not very good at dynamic registration at all */
#define DIGEST_HMAC 666
DEFINE_DIGEST(HMAC, hmac, "hmac", 20);

#ifdef MODULE
#define init_hmac init_module
#endif

int __init init_hmac (void)
{
    /*
     * XXX dorky.  Instead we should be building
     * a dynamic digest_implementation based on 
     * the underlying digest given as a module
     * parameter, but that looks hard given the
     * increadbily static-biased API 
     * in crypto.h, so I'll do it later.
     *
     * for now you only get one digest per 
     * module load.  lame.
     */
    char *digest = "sha1";

    global_H = find_digest_by_name(digest);
    if(global_H == NULL) 
        return -ENXIO; 

	register_digest (&hmac);

    global_H->lock();

	return 0;
}

#ifdef MODULE
void cleanup_module (void)
{
	unregister_digest (&hmac);
    global_H->unlock();
}
#endif
